#include "DL.hpp"  

/////////////////////////////////

// Fonctions associées à la classe DL2

std::ostream& operator<<(std::ostream &flux, DL2 const& f)
{
  flux << f.f0; // On affiche le coefficient constant

  if (f.f1 == 0) // Pour un développement constant on écrit "a + o(x)"
	{
	  flux << " + o(x)";
	  return flux;
	}

  // Si le coefficient en x vaut 1 ou -1, on écrit simplement "x" ou "-x"
  if (f.f1 == 1)
	return flux << " + x + o(x)";

  if (f.f1 == -1)
	return flux << " - x + o(x)";

  // Sinon, on sépare en fonction du signe du coefficient
  if (f.f1>0)
	flux << " + " << f.f1;
  else
	flux << " - " << -f.f1;

  flux << "x + o(x)";
  
  return flux;
}

// Somme, calculée coefficient par coefficient
DL2 operator+(DL2 const& f, DL2 const& g)
{
  DL2 somme = f;
  somme.f0 += g.f0;
  somme.f1 += g.f1;
  return somme;
}

// Différence, calculée coefficient par coefficient
DL2 operator-(DL2 const& f, DL2 const& g)
{
  DL2 diff = f;
  diff.f0 -= g.f0;
  diff.f1 -= g.f1;
  return diff;
}

// Produit, calculée en développant
DL2 operator*(DL2 const& f, DL2 const& g)
{
  DL2 prod = f;
  prod.f0 *= g.f0;
  prod.f1 = f.f0*g.f1 + f.f1*g.f0;
  return prod;
}

// Quotient, qui n'a desens que pour g.f0 != 0
DL2 operator/(DL2 const& f, DL2 const& g)
{
  DL2 ratio = f;
  ratio.f0 /= g.f0;
  ratio.f1 = (f.f1*g.f0 - f.f0*g.f1) / (g.f0*g.f0);
  return ratio;
}



///////////////////////////////////////////////////////

// Fonctions associées à la classe DL

std::ostream& operator<<(std::ostream &flux, DL const& f)
{
  // On affiche les coefficients non nuls un par un (ici, on pourra avoir un affichage peu esthétique comme "x + -3x^2" ; il faudrait détailler un peu plus si on voulait afficher "x - 3x^2")
  if (f.coefs[0] != 0)
	flux << f.coefs[0] << " + ";
  
  for (int i=1; i<=f.ordre; i++)
	if (f.coefs[i] != 0)
	  flux << f.coefs[i] << "x^" << i << " + ";

  flux << "o(x^" << f.ordre << ")";
    
  return flux;
}

// Somme, calculée coefficient par coefficient
// Il faut initialement faire attention au terme négligeable de chaque développement ( par exemple (x+o(x^3)) + (x+x^4+o(x^4)) vaut (2x+o(x^3)))
DL operator+(DL const& f, DL const& g)
{
  DL const& max = (f.ordre>g.ordre) ? f : g;
  DL const& min = (f.ordre>g.ordre) ? g : f;

  DL somme = min; // On initialise la variable "somme" au développement de pus faible ordre.
  for (int i=0; i<=min.ordre; i++)
	somme.coefs[i] += max.coefs[i];

  return somme;
}

// Différence, calculée coefficient par coefficient
// Il faut encore faire attention aux termes négligeables
DL operator-(DL const& f, DL const& g)
{
  DL const& max = (f.ordre>g.ordre) ? f : g;
  DL const& min = (f.ordre>g.ordre) ? g : f;

  DL somme = min;
  for (int i=0; i<=min.ordre; i++)
	somme.coefs[i] -= min.coefs[i];

  return somme;
}

// Pour le produit, on doit faire encore plus attention à l'ordre
// (par exemple (x+o(x^4))*(x^2+o(x^2)) vaut x^3+o(x^3)
DL operator*(DL const& f, DL const& g)
{
  //On cherche les premiers coefficients non nuls de f et de g.
  int min_f=0;
  while (f.coefs[min_f]==0 and min_f<=f.ordre)
	min_f++;

  int min_g=0;
  while (g.coefs[min_g]==0 and min_g<=g.ordre)
	min_g++;

  // La variable "ordre" contient l'ordre du produit
  int ordre = min_g + f.ordre;
  if (min_f + g.ordre < ordre)
	ordre = min_f + g.ordre;

  //ON crée une variable "produit" qui contiendra les coefficients du produit.
  DL prod(0,ordre);
  for (int i=0; i<=ordre; i++)
	for (int k=0; k<=i; k++)
	  if (k<=f.ordre and i-k<=g.ordre)
		prod.coefs[i] += f.coefs[k]*g.coefs[i-k]; // On calcule un par un les coefficients
  
  return prod;
}


// Pour calculer la composée de deux développements
DL DL::operator()(DL const&f)
{
  // On calcule a0 + a1*f + a2*f^2 + a3*f^4... en sommant souccessivement les puissances de f.
  DL fn = f;
  DL composee = coefs[0] + coefs[1]*fn;
  
  for (int i=2; i<=ordre; i++)
	{
	  fn = fn*f;
	  composee = composee + coefs[i]*fn;
	}
  return composee;
}


// Le quation, qui n'existe que pour g.coefs[0] != 0
DL operator/(DL const& f, DL const& g)
{
  DL u = f;
  DL v = g;

  double a = g.coefs[0];
  
  // On réécrit f/g sous la forme u/(1 - v). L'inverse de (1-v) est alors donné par 1+v+v^2+v^3...
  for (int i=0; i<=f.ordre; i++)
	u.coefs[i] /= a;

  v.coefs[0] = 0;
  for (int i=1; i<=g.ordre; i++)
	  v.coefs[i] /= -a;

  
  std::vector<double> inv(g.ordre+1,1);
  DL Inv(inv);// Inv est le développement 1 + x + x^2 +...

  // On renvoie f/g = u/(1-v) = u*(1+v+v^2+...)
  return u * Inv(v);
}
